package dev.shadowsoffire.apotheosis.ench.library;

import dev.shadowsoffire.apotheosis.Apoth;
import net.minecraft.network.protocol.Packet;
import net.minecraftforge.common.capabilities.ForgeCapabilities;
import net.minecraft.core.Direction;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraft.network.protocol.game.ClientboundBlockEntityDataPacket;
import net.minecraft.network.Connection;
import net.minecraft.resources.ResourceLocation;
import it.unimi.dsi.fastutil.objects.ObjectIterator;
import net.minecraft.nbt.Tag;
import net.minecraftforge.registries.ForgeRegistries;
import net.minecraft.nbt.CompoundTag;
import java.util.Iterator;
import dev.shadowsoffire.placebo.recipe.VanillaPacketDispatcher;
import java.util.Map;
import net.minecraft.world.item.enchantment.EnchantmentHelper;
import net.minecraft.world.item.Items;
import net.minecraft.world.item.ItemStack;
import java.util.HashSet;
import it.unimi.dsi.fastutil.objects.Object2IntOpenHashMap;
import net.minecraft.world.level.block.state.BlockState;
import net.minecraft.core.BlockPos;
import net.minecraft.world.level.block.entity.BlockEntityType;
import net.minecraftforge.items.IItemHandler;
import net.minecraftforge.common.util.LazyOptional;
import java.util.Set;
import net.minecraft.world.item.enchantment.Enchantment;
import it.unimi.dsi.fastutil.objects.Object2IntMap;
import net.minecraft.world.level.block.entity.BlockEntity;

public abstract class EnchLibraryTile extends BlockEntity
{
    protected final Object2IntMap<Enchantment> points;
    protected final Object2IntMap<Enchantment> maxLevels;
    protected final Set<EnchLibraryContainer> activeContainers;
    protected final LazyOptional<IItemHandler> itemHandler;
    protected final int maxLevel;
    protected final int maxPoints;
    
    public EnchLibraryTile(final BlockEntityType<?> type, final BlockPos pos, final BlockState state, final int maxLevel) {
        super((BlockEntityType)type, pos, state);
        this.points = (Object2IntMap<Enchantment>)new Object2IntOpenHashMap();
        this.maxLevels = (Object2IntMap<Enchantment>)new Object2IntOpenHashMap();
        this.activeContainers = new HashSet<EnchLibraryContainer>();
        this.itemHandler = (LazyOptional<IItemHandler>)LazyOptional.of(() -> new EnchLibItemHandler());
        this.maxLevel = maxLevel;
        this.maxPoints = levelToPoints(maxLevel);
    }
    
    public void depositBook(final ItemStack book) {
        if (book.m_41720_() != Items.f_42690_) {
            return;
        }
        final Map<Enchantment, Integer> enchs = EnchantmentHelper.m_44831_(book);
        for (final Map.Entry<Enchantment, Integer> e : enchs.entrySet()) {
            if (e.getKey() != null) {
                if (e.getValue() == null) {
                    continue;
                }
                int newPoints = Math.min(this.maxPoints, this.points.getInt((Object)e.getKey()) + levelToPoints(e.getValue()));
                if (newPoints < 0) {
                    newPoints = this.maxPoints;
                }
                this.points.put((Object)e.getKey(), newPoints);
                this.maxLevels.put((Object)e.getKey(), Math.min(this.maxLevel, Math.max(this.maxLevels.getInt((Object)e.getKey()), e.getValue())));
            }
        }
        if (enchs.size() > 0) {
            VanillaPacketDispatcher.dispatchTEToNearbyPlayers((BlockEntity)this);
        }
        this.m_6596_();
    }
    
    public void extractEnchant(final ItemStack stack, final Enchantment ench, final int level) {
        final int curLvl = EnchantmentHelper.m_44831_(stack).getOrDefault(ench, 0);
        if (stack.m_41619_() || !this.canExtract(ench, level, curLvl) || level == curLvl) {
            return;
        }
        final Map<Enchantment, Integer> enchs = EnchantmentHelper.m_44831_(stack);
        enchs.put(ench, level);
        EnchantmentHelper.m_44865_((Map)enchs, stack);
        this.points.put((Object)ench, Math.max(0, this.points.getInt((Object)ench) - levelToPoints(level) + levelToPoints(curLvl)));
        if (!this.f_58857_.m_5776_()) {
            VanillaPacketDispatcher.dispatchTEToNearbyPlayers((BlockEntity)this);
        }
        this.m_6596_();
    }
    
    public boolean canExtract(final Enchantment ench, final int level, final int currentLevel) {
        return this.maxLevels.getInt((Object)ench) >= level && this.points.getInt((Object)ench) >= levelToPoints(level) - levelToPoints(currentLevel);
    }
    
    public static int levelToPoints(final int level) {
        return (int)Math.pow(2.0, level - 1);
    }
    
    public void m_183515_(final CompoundTag tag) {
        final CompoundTag points = new CompoundTag();
        for (final Object2IntMap.Entry<Enchantment> e : this.points.object2IntEntrySet()) {
            points.m_128405_(ForgeRegistries.ENCHANTMENTS.getKey((Object)e.getKey()).toString(), e.getIntValue());
        }
        tag.m_128365_("Points", (Tag)points);
        final CompoundTag levels = new CompoundTag();
        for (final Object2IntMap.Entry<Enchantment> e2 : this.maxLevels.object2IntEntrySet()) {
            levels.m_128405_(ForgeRegistries.ENCHANTMENTS.getKey((Object)e2.getKey()).toString(), e2.getIntValue());
        }
        tag.m_128365_("Levels", (Tag)levels);
        super.m_183515_(tag);
    }
    
    public void m_142466_(final CompoundTag tag) {
        super.m_142466_(tag);
        final CompoundTag points = tag.m_128469_("Points");
        for (final String s : points.m_128431_()) {
            final Enchantment ench = (Enchantment)ForgeRegistries.ENCHANTMENTS.getValue(new ResourceLocation(s));
            if (ench == null) {
                continue;
            }
            this.points.put((Object)ench, points.m_128451_(s));
        }
        final CompoundTag levels = tag.m_128469_("Levels");
        for (final String s2 : levels.m_128431_()) {
            final Enchantment ench2 = (Enchantment)ForgeRegistries.ENCHANTMENTS.getValue(new ResourceLocation(s2));
            if (ench2 == null) {
                continue;
            }
            this.maxLevels.put((Object)ench2, levels.m_128451_(s2));
        }
    }
    
    public void onDataPacket(final Connection net, final ClientboundBlockEntityDataPacket pkt) {
        final CompoundTag tag = pkt.m_131708_();
        final CompoundTag points = tag.m_128469_("Points");
        for (final String s : points.m_128431_()) {
            final Enchantment ench = (Enchantment)ForgeRegistries.ENCHANTMENTS.getValue(new ResourceLocation(s));
            if (ench == null) {
                continue;
            }
            this.points.put((Object)ench, points.m_128451_(s));
        }
        final CompoundTag levels = tag.m_128469_("Levels");
        for (final String s2 : levels.m_128431_()) {
            final Enchantment ench2 = (Enchantment)ForgeRegistries.ENCHANTMENTS.getValue(new ResourceLocation(s2));
            if (ench2 == null) {
                continue;
            }
            this.maxLevels.put((Object)ench2, levels.m_128451_(s2));
        }
        this.activeContainers.forEach(EnchLibraryContainer::onChanged);
    }
    
    public ClientboundBlockEntityDataPacket getUpdatePacket() {
        return ClientboundBlockEntityDataPacket.m_195640_((BlockEntity)this);
    }
    
    public CompoundTag m_5995_() {
        final CompoundTag tag = super.m_5995_();
        final CompoundTag points = new CompoundTag();
        for (final Object2IntMap.Entry<Enchantment> e : this.points.object2IntEntrySet()) {
            points.m_128405_(ForgeRegistries.ENCHANTMENTS.getKey((Object)e.getKey()).toString(), e.getIntValue());
        }
        tag.m_128365_("Points", (Tag)points);
        final CompoundTag levels = new CompoundTag();
        for (final Object2IntMap.Entry<Enchantment> e2 : this.maxLevels.object2IntEntrySet()) {
            levels.m_128405_(ForgeRegistries.ENCHANTMENTS.getKey((Object)e2.getKey()).toString(), e2.getIntValue());
        }
        tag.m_128365_("Levels", (Tag)levels);
        return tag;
    }
    
    public Object2IntMap<Enchantment> getPointsMap() {
        return this.points;
    }
    
    public Object2IntMap<Enchantment> getLevelsMap() {
        return this.maxLevels;
    }
    
    public void addListener(final EnchLibraryContainer ctr) {
        this.activeContainers.add(ctr);
    }
    
    public void removeListener(final EnchLibraryContainer ctr) {
        this.activeContainers.remove(ctr);
    }
    
    public int getMax(final Enchantment ench) {
        return Math.min(this.maxLevel, this.maxLevels.getInt((Object)ench));
    }
    
    public <T> LazyOptional<T> getCapability(final Capability<T> cap, final Direction side) {
        if (cap == ForgeCapabilities.ITEM_HANDLER) {
            return (LazyOptional<T>)this.itemHandler.cast();
        }
        return (LazyOptional<T>)super.getCapability((Capability)cap, side);
    }
    
    private class EnchLibItemHandler implements IItemHandler
    {
        public int getSlots() {
            return 1;
        }
        
        public ItemStack getStackInSlot(final int slot) {
            return ItemStack.f_41583_;
        }
        
        public ItemStack insertItem(final int slot, final ItemStack stack, final boolean simulate) {
            if (stack.m_41720_() != Items.f_42690_ || stack.m_41613_() > 1) {
                return stack;
            }
            if (!simulate) {
                EnchLibraryTile.this.depositBook(stack);
            }
            return ItemStack.f_41583_;
        }
        
        public ItemStack extractItem(final int slot, final int amount, final boolean simulate) {
            return ItemStack.f_41583_;
        }
        
        public int getSlotLimit(final int slot) {
            return 1;
        }
        
        public boolean isItemValid(final int slot, final ItemStack stack) {
            return slot == 0 && stack.m_41720_() == Items.f_42690_;
        }
    }
    
    public static class BasicLibraryTile extends EnchLibraryTile
    {
        public BasicLibraryTile(final BlockPos pos, final BlockState state) {
            super((BlockEntityType<?>)Apoth.Tiles.LIBRARY.get(), pos, state, 16);
        }
    }
    
    public static class EnderLibraryTile extends EnchLibraryTile
    {
        public EnderLibraryTile(final BlockPos pos, final BlockState state) {
            super((BlockEntityType<?>)Apoth.Tiles.ENDER_LIBRARY.get(), pos, state, 31);
        }
    }
}
